#
# Copyright (C) 2021 Supercomputing Systems AG
# This file is part of smartmeter-datacollector.
#
# SPDX-License-Identifier: GPL-2.0-only
# See LICENSES/README.md for more information.
#
import logging
from typing import List

import serial

from .cosem import CosemConfig, RegisterCosem
from .hdlc_dlms_parser import HdlcDlmsParser
from .reader import Reader
from .reader_data import ReaderDataPoint, ReaderDataPointTypes
from .serial_reader import SerialConfig, SerialReader

LOGGER = logging.getLogger("smartmeter")

LGE450_COSEM_REGISTERS = [
    RegisterCosem("1.0.1.7.0.255", ReaderDataPointTypes.ACTIVE_POWER_P.value),
    RegisterCosem("1.0.2.7.0.255", ReaderDataPointTypes.ACTIVE_POWER_N.value),
    RegisterCosem("1.0.3.7.0.255", ReaderDataPointTypes.REACTIVE_POWER_P.value),
    RegisterCosem("1.0.4.7.0.255", ReaderDataPointTypes.REACTIVE_POWER_N.value),

    RegisterCosem("1.0.13.7.0.255", ReaderDataPointTypes.POWER_FACTOR.value, 0.001),
    RegisterCosem("1.0.14.7.0.255", ReaderDataPointTypes.NET_FREQUENCY.value),

    RegisterCosem("1.0.21.7.0.255", ReaderDataPointTypes.ACTIVE_POWER_P_L1.value),
    RegisterCosem("1.0.22.7.0.255", ReaderDataPointTypes.ACTIVE_POWER_N_L1.value),
    RegisterCosem("1.0.23.7.0.255", ReaderDataPointTypes.REACTIVE_POWER_P_L1.value),
    RegisterCosem("1.0.24.7.0.255", ReaderDataPointTypes.REACTIVE_POWER_N_L1.value),

    RegisterCosem("1.0.31.7.0.255", ReaderDataPointTypes.CURRENT_L1.value),
    RegisterCosem("1.0.32.7.0.255", ReaderDataPointTypes.VOLTAGE_L1.value),

    RegisterCosem("1.0.41.7.0.255", ReaderDataPointTypes.ACTIVE_POWER_P_L2.value),
    RegisterCosem("1.0.42.7.0.255", ReaderDataPointTypes.ACTIVE_POWER_N_L2.value),
    RegisterCosem("1.0.43.7.0.255", ReaderDataPointTypes.REACTIVE_POWER_P_L2.value),
    RegisterCosem("1.0.44.7.0.255", ReaderDataPointTypes.REACTIVE_POWER_N_L2.value),

    RegisterCosem("1.0.51.7.0.255", ReaderDataPointTypes.CURRENT_L2.value),
    RegisterCosem("1.0.52.7.0.255", ReaderDataPointTypes.VOLTAGE_L2.value),

    RegisterCosem("1.0.61.7.0.255", ReaderDataPointTypes.ACTIVE_POWER_P_L3.value),
    RegisterCosem("1.0.62.7.0.255", ReaderDataPointTypes.ACTIVE_POWER_N_L3.value),
    RegisterCosem("1.0.63.7.0.255", ReaderDataPointTypes.REACTIVE_POWER_P_L3.value),
    RegisterCosem("1.0.64.7.0.255", ReaderDataPointTypes.REACTIVE_POWER_N_L3.value),

    RegisterCosem("1.0.71.7.0.255", ReaderDataPointTypes.CURRENT_L3.value),
    RegisterCosem("1.0.72.7.0.255", ReaderDataPointTypes.VOLTAGE_L3.value),

    RegisterCosem("1.0.81.7.40.255", ReaderDataPointTypes.ANGLE_UI_L1.value),
    RegisterCosem("1.0.81.7.51.255", ReaderDataPointTypes.ANGLE_UI_L2.value),
    RegisterCosem("1.0.81.7.62.255", ReaderDataPointTypes.ANGLE_UI_L3.value),

    RegisterCosem("1.1.1.8.0.255", ReaderDataPointTypes.ACTIVE_ENERGY_P.value),
    RegisterCosem("1.1.2.8.0.255", ReaderDataPointTypes.ACTIVE_ENERGY_N.value),
    RegisterCosem("1.1.3.8.0.255", ReaderDataPointTypes.REACTIVE_ENERGY_P.value),
    RegisterCosem("1.1.4.8.0.255", ReaderDataPointTypes.REACTIVE_ENERGY_N.value),

    RegisterCosem("1.1.5.8.0.255", ReaderDataPointTypes.REACTIVE_ENERGY_Q1.value),
    RegisterCosem("1.1.6.8.0.255", ReaderDataPointTypes.REACTIVE_ENERGY_Q2.value),
    RegisterCosem("1.1.7.8.0.255", ReaderDataPointTypes.REACTIVE_ENERGY_Q3.value),
    RegisterCosem("1.1.8.8.0.255", ReaderDataPointTypes.REACTIVE_ENERGY_Q4.value),
]


class LGE450(Reader):
    HDLC_FLAG = b"\x7e"

    def __init__(self, port: str, decryption_key: str = None) -> None:
        super().__init__()

        serial_config = SerialConfig(
            port=port,
            baudrate=2400,
            data_bits=serial.EIGHTBITS,
            parity=serial.PARITY_EVEN,
            stop_bits=serial.STOPBITS_ONE,
            termination=LGE450.HDLC_FLAG
        )
        self._serial = SerialReader(serial_config, self._data_received)
        cosem_config = CosemConfig(
            id_obis="0.0.42.0.0.255",
            clock_obis="0.0.1.0.0.255",
            register_obis=LGE450_COSEM_REGISTERS
        )
        self._parser = HdlcDlmsParser(cosem_config, decryption_key)

    async def start(self) -> None:
        await self._serial.start_and_listen()

    def _data_received(self, received_data: bytes) -> None:
        if not received_data:
            return
        if received_data == LGE450.HDLC_FLAG:
            self._parser.append_to_hdlc_buffer(received_data)
            return

        self._parser.append_to_hdlc_buffer(received_data)
        if self._parser.extract_data_from_hdlc_frames():
            dlms_objects = self._parser.parse_to_dlms_objects()
            data_points = self._parser.convert_dlms_bundle_to_reader_data(dlms_objects)
            self._forward_data_points(data_points)

    def _forward_data_points(self, data_points: List[ReaderDataPoint]) -> None:
        self._notify_observers(data_points)



#
# Copyright (C) 2021 Supercomputing Systems AG
# This file is part of smartmeter-datacollector.
#
# SPDX-License-Identifier: GPL-2.0-only
# See LICENSES/README.md for more information.
#
import logging

import serial

from .cosem import CosemConfig
from .hdlc_dlms_parser import HdlcDlmsParser
from .reader import Reader
from .serial_reader import SerialConfig, SerialReader

LOGGER = logging.getLogger("smartmeter")


class IskraAM550(Reader):
    HDLC_FLAG = b"\x7e"

    def __init__(self, port: str) -> None:
        super().__init__()

        serial_config = SerialConfig(
            port=port,
            baudrate=115200,
            data_bits=serial.EIGHTBITS,
            parity=serial.PARITY_NONE,
            stop_bits=serial.STOPBITS_ONE,
            termination=IskraAM550.HDLC_FLAG
        )
        self._serial = SerialReader(serial_config, self._data_received)
        cosem_config = CosemConfig(
            id_obis="0.0.42.0.0.255",  # TODO: set correct OBIS
            clock_obis="0.0.1.0.0.255",  # TODO: set correct OBIS
            register_obis=[]
        )
        self._parser = HdlcDlmsParser(cosem_config)

    async def start(self) -> None:
        await self._serial.start_and_listen()

    def _data_received(self, received_data: bytes) -> None:
        if not received_data:
            return
        if received_data == self.HDLC_FLAG:
            self._parser.append_to_hdlc_buffer(received_data)
            return
        self._parser.append_to_hdlc_buffer(received_data)
        if self._parser.extract_data_from_hdlc_frames():
            dlms_objects = self._parser.parse_to_dlms_objects()
            LOGGER.debug(dlms_objects)



#
# Copyright (C) 2021 Supercomputing Systems AG
# This file is part of smartmeter-datacollector.
#
# SPDX-License-Identifier: GPL-2.0-only
# See LICENSES/README.md for more information.
#
from dataclasses import dataclass
from typing import List, Optional

from .reader_data import ReaderDataPointType


@dataclass
class RegisterCosem:
    obis: str
    data_point_type: ReaderDataPointType
    scaling: float = 1.0


class CosemConfig:
    def __init__(self, id_obis: str, clock_obis: str, register_obis: List[RegisterCosem]) -> None:

        self._id_obis = id_obis
        self._clock_obis = clock_obis
        self._register_obis = {r.obis: r for r in register_obis}

    @property
    def id_obis(self) -> str:
        return self._id_obis

    @property
    def clock_obis(self) -> str:
        return self._clock_obis

    def get_register(self, obis: str) -> Optional[RegisterCosem]:
        return self._register_obis.get(obis, None)



#
# Copyright (C) 2021 Supercomputing Systems AG
# This file is part of smartmeter-datacollector.
#
# SPDX-License-Identifier: GPL-2.0-only
# See LICENSES/README.md for more information.
#
from abc import ABC, abstractmethod
from typing import List

from .reader_data import ReaderDataPoint


class Reader(ABC):
    def __init__(self) -> None:
        self._observers = []

    def register(self, observer) -> None:
        self._observers.append(observer)

    @abstractmethod
    async def start(self) -> None:
        raise NotImplementedError()

    def _notify_observers(self, data_points: List[ReaderDataPoint]) -> None:
        for observer in self._observers:
            observer.notify(data_points)



#
# Copyright (C) 2021 Supercomputing Systems AG
# This file is part of smartmeter-datacollector.
#
# SPDX-License-Identifier: GPL-2.0-only
# See LICENSES/README.md for more information.
#
from dataclasses import dataclass

import serial
from aioserial import AioSerial


@dataclass
class SerialConfig:
    port: str
    baudrate: int = 9600
    data_bits: int = serial.EIGHTBITS
    parity: str = serial.PARITY_NONE
    stop_bits: int = serial.STOPBITS_ONE
    termination: bytes = serial.LF


# pylint: disable=too-few-public-methods
class SerialReader:
    CHUNK_SIZE = 16

    def __init__(self, serial_config: SerialConfig, callback) -> None:
        self._callback = callback
        self._termination = serial_config.termination
        self._serial = AioSerial(
            port=serial_config.port,
            baudrate=serial_config.baudrate,
            bytesize=serial_config.data_bits,
            parity=serial_config.parity,
            stopbits=serial_config.stop_bits
        )

    async def start_and_listen(self) -> None:
        while True:
            data: bytes = await self._serial.read_until_async(self._termination, None)
            self._callback(data)
